import tonkin.values

# Given a series of bytes representing a packet and
# a list of variable dicts, return a list of bytestrings
# representing variable values for the packet
def _splitPacket(rawPacket, vars):
	splitPacket = []
	for var in vars:
		splitPacket.append(rawPacket[:var["Length"]])
		rawPacket = rawPacket[var["Length"]:]
		
	return splitPacket

# Given a series of bytes representing a packet,
# return a dictionary representing the packet
def readRawPacket(rawPacket, vars):
	vals = _splitPacket(rawPacket, vars)
	packet = {}
	for index, var in enumerate(vars):
		varName = var["Name"]
		value = tonkin.values.readRawValue(vals[index], var)
		packet[varName] = value
	return packet



import tonkin.arrays

def _calculatePacketLength(vars):
	length = 0
	for var in vars:
		length += var["Length"]
	return length

# Given a data stream and the list of variable types,
# get a list of packets from the data stream
def splitStreamToPackets(stream, vars):
	packetLength = _calculatePacketLength(vars)
	return tonkin.arrays.splitArrayIntoLengths(stream, packetLength)



import tonkin.fileops
import countach
import tonkin.wholedata
import tonkin.variables
import tonkin.dataStream
import tonkin.packets
import csv

# Given a dat file and an a2l file, returns a list of dicts
# each representing a single packet
def readDatFileWithA2L(datFile, a2lFile):
	# Import the files
	a2lData = countach.extractData(a2lFile)
	fileBytes = tonkin.fileops.importFile(datFile)
	# Split the dat file up into sections
	header, vars, stream = tonkin.wholedata.splitFileData(fileBytes)
	# Extract the variables from the variable section
	variableList = tonkin.variables.getVariableListFromA2L(vars, a2lData)
	rawPackets = tonkin.dataStream.splitStreamToPackets(stream, variableList)
	output = []
	variableList.reverse()
	for packet in rawPackets:
		output.append(tonkin.packets.readRawPacket(packet, variableList))
	return output

def datFileTo2DArrayA2L(datFile, a2lFile):
	data = readDatFileWithA2L(datFile, a2lFile)
	output = []
	
	# Get headers
	headers = []
	for key in data[0]:
		headers.append(key)
	output.append(headers)

	
	for packet in data:
		row = []
		for key in packet:
			row.append(packet[key])
		output.append(row)
	return output

def datFileToCSVWithA2L(datFile, a2lFile, csvFile, delimiter=","):
	twoD = datFileTo2DArrayA2L(datFile, a2lFile)
	with open(csvFile, "w+") as f:
		csvWriter = csv.writer(f, delimiter=delimiter)
		csvWriter.writerows(twoD)



def _checkIfFixedIsSigned(typeString):
	if typeString[0] == "u":
		signed = 0
	elif typeString[0] == "s":
		signed = 1
	else:
		raise ValueError("Sign indication not found in type string")
	return signed

def _getWordLength(typeString):
	# Remove unneccessary data
	typeString = typeString[4:]
	typeString = typeString.split("_")
	return int(typeString[0])

# TODO: Check if negative total slopes possible?
def _getTotalSlope(typeString):
	totalSlope = 1
	if "S" in typeString:
		totalSlopeIndex = typeString.index("S") + 1
		totalSlope = int(typeString[totalSlopeIndex])
	elif "E" in typeString:
		totalSlopeIndex = typeString.index("E") + 1
		totalSlope = 2 ** int(typeString[totalSlopeIndex])
	return totalSlope

# TODO: Check if negative biases possible?
def _getBias(typeString):
	bias = 0
	if "B" in typeString:
		biasIndex = typeString.index("B") + 1
		bias = int(typeString[biasIndex])
	return bias

def getFixedType(typeString):
	signed = _checkIfFixedIsSigned(typeString)
	wordLength = _getWordLength(typeString)
	totalSlope = _getTotalSlope(typeString)
	bias = _getBias(typeString)
	return signed, wordLength, totalSlope, bias
	
# Round n up to the nearest multiple
def roundUp(n, multiple):
	if multiple == 0:
		return n
	
	remainder = n % multiple
	
	if remainder == 0:
		return n

	return n + multiple - remainder

def getByteLength(bitLength):
	totalBits = roundUp(bitLength, 8)
	return totalBits / 8

def getByteLengthFromName(typeName):
	_, wordLength, _, _ = getFixedType(typeName)
	return getByteLength(wordLength)

# Remove the extra bits from the start of a fixdt value
def chopExtraBits(value, wordLength, byteLength):
	# The case where the wordLength is a multiple of 8
	if wordLength / 8 == byteLength:
		return value
	
	bitsToChop = (byteLength * 8) - wordLength
	binary = bin(value)
	# Chop an extra 2 chars for "0b"
	chopped = binary[bitsToChop + 2:]
	return int(chopped)



import struct
import tonkin.fixed

def _convertint8(int8B):
	return int.from_bytes(int8B, "little", signed=True)

def _convertint16(int16B):
	return int.from_bytes(int16B, "little", signed=True)

def _convertint32(int32B):
	return int.from_bytes(int32B, "little", signed=True)

def _convertuint8(uint8B):
	return int.from_bytes(uint8B, "little", signed=False)

def _convertuint16(uint16B):
	return int.from_bytes(uint16B, "little", signed=False)

def _convertSingle(singleB):
	return struct.unpack("f", singleB)[0]

def _convertBoolean(boolB):
	if boolB == b'\x01':
		return True
	else:
		return False

def _convertFixed(fixedB, rawVarType):
	signed, length, _, _ = tonkin.fixed.getFixedType(rawVarType)
	rawValue = tonkin.fixed.chopExtraBits(fixedB)
	if signed == 1:
		rawValue -= 2 ** length
	return rawValue

# Given a series of bytes representing a value within a packet
# and a variable dict, return the value of the 
def readRawValue(rawValue, var):
	varType = var["Type"]
	
	if varType == "int8":
		return _convertint8(rawValue)
	elif varType == "int16":
		return _convertint16(rawValue)
	elif varType == "int32":
		return _convertint32(rawValue)
	elif varType == "uint8":
		return _convertuint8(rawValue)
	elif varType == "uint16":
		return _convertuint16(rawValue)
	elif varType == "single":
		return _convertSingle(rawValue)
	elif varType == "boolean":
		return _convertBoolean(rawValue)
	elif "fix" in varType:
		return _convertFixed(rawValue, varType)



def shiftArrayLeft(array, shift):
	# Bring large numbers down to smaller equivalents
	# and handle negatives for shifts right
	shift = shift % len(array)

	for i in range(0, shift):
		array.append(array.pop(0))
	return array

# Given a source array and a target array, returns the index of where the target
# array appears within the source array (returns the index of the final value)
def getSubArrayIndex(source, target):
	# TODO: Ensure that the buffer items don't appear in split
	buffer = [None] * len(target)
	
	for j, i in enumerate(source):
		buffer = shiftArrayLeft(buffer, 1)
		buffer[-1] = i

		if buffer == target:
			return j
	
	raise ValueError("Sub-array not found")

def getSubArrayIndices(source, target):
	output = []
	
	# TODO: Ensure that the buffer items don't appear in split
	buffer = [None] * len(target)
	
	for j, i in enumerate(source):
		buffer = shiftArrayLeft(buffer, 1)
		buffer[-1] = i

		if buffer == target:
			output.append(j)
	
	return output

# TODO: Test for different first elements of source
def splitArrayBySub(source, target):
	output = []
	indices = getSubArrayIndices(source, target)
	
	# The case where the target does not appear
	if indices == []:
		return [source]

	for j, i in enumerate(indices):
		startOfBlock = i - len(target) + 1
		
		if j < len(indices) - 1:
			endOfBlock = indices[j+1] - len(target) + 1
		else:
			endOfBlock = len(source)

		output.append(source[startOfBlock:endOfBlock])

	return output

def splitArrayIntoLengthsGen(array, lengths):
	for i in range(0, len(array), lengths):
		yield array[i:i + lengths]

def splitArrayIntoLengths(array, lengths):
	return list(splitArrayIntoLengthsGen(array, lengths))



import tonkin.arrays

def extractHeaderSec(fileBytes):
	return fileBytes[0:228]

def extractVariablesSec(fileBytes):
	output = fileBytes[228:]
	end = tonkin.arrays.getSubArrayIndex(output, [0x43, 0x47, 0x1A])
	return output[:end]

def extractDataStream(fileBytes, prePrepared=False):
	# Remove the header and variable definition sections
	# if it hasn't been done already
	if prePrepared:
		output = fileBytes
	else:
		headerLength = len(extractHeaderSec(fileBytes))
		variableLength = len(extractVariablesSec(fileBytes))
		upperLength = headerLength + variableLength
		output = fileBytes[upperLength:]

	# The sequence of bytes that marks the start of the data stream
	startMarker = [0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]
	# The index of the first byte of the data stream
	start = tonkin.arrays.getSubArrayIndex(output, startMarker) + 1
	
	return output[start:]

def splitFileData(fileBytes):
	header = extractHeaderSec(fileBytes)
	variables = extractVariablesSec(fileBytes)
	# Chop the upper sections away
	upperLength = len(header) + len(variables)
	fileBytes = fileBytes[upperLength:]
	dataStream = extractDataStream(fileBytes, prePrepared=True)
	return header, variables, dataStream



def verifyASCII(start, stop, result, source):
	return source[start:stop].decode("ASCII") == result

# Check if the file uses the MDF format
def checkFormat(header):
	return verifyASCII(0, 3, "MDF", header)

# Check if the file uses MDF 3.00
def checkVersion(header):
	return verifyASCII(8, 12, "3.00", header)

# Check if the vendor is Ecotrons
def checkVendor(header):
	return verifyASCII(16, 24, "Ecotrons", header)

# Check if the file was created by EcoCAL
def checkSoftware(header):
	return verifyASCII(100, 106, "EcoCAL", header)

# Check that the second vendor is Ecotrons
def checkSecondVendor(header):
	return verifyASCII(132, 140, "Ecotrons", header)

def checkHeader(header):
	if not checkFormat(header):
		return False
	elif not checkVersion(header):
		return False
	elif not checkVendor(header):
		return False
	elif not checkSoftware(header):
		return False
	elif not checkSecondVendor(header):
		return False
	return True



import tonkin.arrays
import data
import fixed

# Splits the variable definition section into individual definitions
def _splitVariables(variableSection):
	return tonkin.arrays.splitArrayBySub(variableSection, [0x43, 0x43, 0x3E])

# Extract the name of a variable from the definition
def _extractVariableName(definition):
	outputBytes = []
	# Remove data before variable name
	definition = definition[66:]

	# Extract name
	for i in definition:
		if i == 0x00:
			break
		outputBytes.append(chr(i))

	return "".join(outputBytes)

# Extract names from a list of variable definitions
def _getVariableNames(splitVars):
	# Don't get the "time" variable
	splitVars.pop(-1)
	output = []
	for var in splitVars:
		output.append(_extractVariableName(var))
	return output

# Get the type of a variable from it's inclusion in an A2L file
def _getTypeFromA2LData(variableName, a2lData):
	for i in a2lData:
		rightCategory = i["Category"] == "Measurement"
		rightName = i["Name"] == variableName
		if rightCategory and rightName:
			# The type is stored in "conversion method" after the string "test_CM_"
			# This removes that prefix
			rawType = i["Conversion Method"]
			varType = rawType[8:]

			return varType
	raise ValueError("Variable " + variableName + " not found in A2L data")

# Take a list of variable names and return a list of variable dicts with types
def _getTypesFromA2LData(variableNames, a2lData):
	output = []
	for variable in variableNames:
		currentVar = {
			"Name": variable,
			"Type": _getTypeFromA2LData(variable, a2lData)
		}
		output.append(currentVar)
	return output

# Given a data type, return it's length in bytes
def getTypeLength(typeName):
	# If the typeName is a dictionary, meaning it's a fixdt
	if "fix" in typeName:
		return fixed.getByteLengthFromName(typeName)
	return data.typeLengths[typeName]

# Given a variable with a type, return the length of the variable in bytes
def _getVariableLength(variable):
	return getTypeLength(variable["Type"])

# Given a list of variable dicts, returns a list of variable dicts with types
def _getVariableLengths(variables):
	output = []
	for var in variables:
		var["Length"] = _getVariableLength(var)
		output.append(var)
	return output

# Given a variable section and data from an A2L file, return a list of dicts
# representing the variables, including types and lengths in bytes
def getVariableListFromA2L(variableSection, a2lData):
	split = _splitVariables(variableSection)
	names = _getVariableNames(split)
	namesWithTypes = _getTypesFromA2LData(names, a2lData)
	namesWithTypesAndLengths = _getVariableLengths(namesWithTypes)
	# Add the "time" variable present in all files
	namesWithTypesAndLengths.append({"Name": "time", "Type": "single", "Length": 4})
	return namesWithTypesAndLengths
